﻿var fs = require('fs');
var Promise = require('promise');
var jsMapper = require("./js_diff/index.js").javascriptHandler;
var jsMapperV2 = require("./js_diff/index.js").javascriptHandlerV2;
var mapSingleFileInMemory = require("./js_diff/index.js").mapSingleFileInMemory;
var filesMapping = require("./js_diff/files-mapping");
var consoleFileLogger = require("./sl-logger.js").consoleFileLogger;
var version = require("../package.json").version;
var slNodeCover = require('sl-node-cover');
var md5 = require('md5');
var utils = require('./utils.js');
var os = require("os");
var SlJsInfra = require('sl-js-infra').SlJsInfra;
var CiaEnvVars = SlJsInfra.SlEnvVars.CIA;
var DefaultValuesResolver = require('./default-values-resolver');
var fileAndFolderUtils = require("./file-and-folder-utils");

function BuildDiffProcess(cfg, buildMappingProxy, sourceConrolProviders, parsedDependencies, logger) {

    var log = logger ? logger.child({
        className: "BuildDiffProcess"
    }) : consoleFileLogger("BuildDiffProcess");

    function createBuildDiffPromiseChain(buildArguments) {
        var handlerParams = {
            buildArguments: buildArguments,
            cfg: cfg,
            finalOutput: {
                "meta": {
                    "generated": new Date().valueOf(),
                    "build": buildArguments.build,
                    "commit": buildArguments.commit,
                    "customerId": cfg.customerId,
                    "appName": buildArguments.appname,
                    //"technologies": "nodejs",
                    "technology": "nodejs",
                    "branch": buildArguments.branch,
                    "environment": {
                        agentVersion: version,
                        agentType: "nodeJsBuildScanner",
                        processArgv:process.argv,
                        machineName: os.hostname(),
                        platform: os.platform(),
                        os: os.type(),
                        osVersion: os.release(),
                        arch: os.arch(),
                        processId: process.pid,
                        ipAddress: clone(os.networkInterfaces()),
                        dependencies: clone(process.versions),
                        runtime: process.version
                    },
                    "authors": buildArguments.author,
                    "logsUrl": buildArguments.logsUrl,
                    "jobName": buildArguments.jobName,
                    "moduleName":  buildArguments.uniqueModuleId, //Added for backward compatibility incase the server needs it.
                    "uniqueModuleId": buildArguments.uniqueModuleId,
                    "configurationData": buildArguments,
                    "scmProvider": buildArguments.scmProvider,
                    "scmVersion": buildArguments.scmVersion,
                    "scmBaseUrl": buildArguments.scmBaseUrl
                },
                version: "2.0",
                files: [],
                dependencies: parsedDependencies || []
            }
        };

        var scmInstance;

        if (buildArguments.workspacepath) {
            var scmType = sourceConrolProviders[buildArguments.scm];
            if (scmType) {
                scmInstance = new scmType(cfg, buildArguments.workspacePath);
                handlerParams.finalOutput.meta.scm = buildArguments.scm;
                log.info("Created an scmInstance of type '" + scmType.name + "'. Searching for a local repository at '" + buildArguments.workspacepath + "'.");
            }else{
                log.warn("Unsupported scmType: '" + scmType + "'.");
            }
        }

        if (!buildArguments.commit && scmInstance && !scmInstance.detectCommitVersion) {
            return Promise.reject("'commit' command-line argument is required, since automatic detection by the " + buildArguments.scm + " SCM is not supported");
        }

        var promiseChain = Promise.resolve();

        if (buildArguments.workspacepath) {

            //2. If there is a mapping, get the diffs
            if (scmInstance) {
                //2.1 Get repository url from the SCM
                promiseChain = promiseChain.then(function (mapping) {
                    log.info("Getting repository URL.");
                    return getRepositoryUrl(handlerParams, scmInstance);
                });

                //2.2 Get all contributors from the scm
                if(buildArguments.sendContributors){
                    promiseChain = promiseChain.then(function(mapping){
                        log.info("Getting contributors.");
                        return getContributors(handlerParams, scmInstance);
                    });
                }

                //2.3 Get branch history from the SCM
                promiseChain = promiseChain.then(function (mapping) {
                    log.info("Loading branch history.");
                    return loadBranchHistory(handlerParams, scmInstance);
                });

                //2.4 Get all commits effected each file
                if(buildArguments.sendContributors){
                    promiseChain= promiseChain.then(function(mapping){
                        log.info("Getting commits per file.");
                        return getCommitsPerFile(handlerParams, scmInstance);
                    });
                }
                //2.5 if commit was not specified, automatically detect it
                if (!buildArguments.commit) {
                    promiseChain = promiseChain.then(function () {
                        log.info("Couldn't find a 'commit' on the 'buildArguments'. Trying to detect the commit version.");
                        return scmInstance.detectCommitVersion(handlerParams, log).then(function (commit) {
                            if (commit) {
                                handlerParams.finalOutput.meta.commit = commit;
                            }
                        });
                    });
                };
            }

            //3. Call JS mapping handler
            promiseChain = promiseChain.then(function () {
                log.info("Handling files and sending information in the new Build Scanner Format (with UniqueID).");
                var filesMapper = new filesMapping(handlerParams, log);
                return filesMapper.process();
            });
        };

        //4. Submit the final output to the server
        promiseChain = promiseChain.then(function () {
            if (buildArguments.instrumentationOnly) {
                log.info("No need to submit the files to server as the 'instrumentationOnly' flag is on.");
                return Promise.resolve();
            } else {
                //fs.writeFileSync("buildMapping.json", JSON.stringify(handlerParams.finalOutput,null,'\t'));
                //return Promise.resolve();
                return buildMappingProxy.submitBuildMapping(handlerParams.finalOutput);
            }
        });

         //5. Print the counters
         promiseChain = promiseChain.then(function () {
            var mapping = handlerParams.finalOutput;
            if(mapping != null && mapping.counters !=null  && mapping.counters.methods != null && mapping.counters.branches != null){
                console.log("Reported " + mapping.counters.methods + " methods and " + mapping.counters.branches + " branches in " + mapping.files.length + " files.");
            }
        });

        //6. Create instrumented browser files.
        if (buildArguments.instrumentForBrowsers) {
            promiseChain = promiseChain.then(function () {

                if (handlerParams.finalOutput && handlerParams.finalOutput.files) {
                    delete handlerParams.finalOutput.files;
                }
                var promise = new Promise(function (resolve, reject) {
                    var resolveWithoutHash = true;
                    var delayShutdownInSeconds =buildArguments.delayShutdownInSeconds || 30;
                    var downloadAgent = buildArguments.downloadAgent != null ? buildArguments.downloadAgent : true;

                    var instrumentationConfig = {
                            appName: buildArguments.appname,
                            build: buildArguments.build,
                            customerId: cfg.customerId,
                            branch: buildArguments.branch,
                            sourceRoot: buildArguments.workspacepath,
                            files: getProcessedFiles(buildArguments.workspacepath, handlerParams.processedFiles, logger),
                            outputPath: buildArguments.outputpath,
                            copyAllFilesToOutput: buildArguments.copyAllFilesToOutput !== false,
                            buildsessionid: buildArguments.buildsessionid,
                            esModules: buildArguments["es-modules"],
                            server: handlerParams.cfg.server,
                            token: cfg.token,
                            instrumentationType: "browser",
                            resolveWithoutHash: resolveWithoutHash,
                            delayShutdownInSeconds: delayShutdownInSeconds,
                            downloadAgent: downloadAgent,
                            workspacepath: buildArguments.workspacepath
                        };

                    log.info("'instrumentForBrowsers' flag is on. Starting instrumentation. Instrumentation config: ", JSON.stringify(instrumentationConfig));
                    var callback =  function (err) {
                        if (err){
                            log.info("Finished the instrumentation with an error. Error:", err);
                            reject(err);
                        }
                        else{
                            log.info("Finished the instrumentation successfully");
                            return resolve();
                        }
                    };

                    if (CiaEnvVars.isEnableModernInstrumentation()){
                        doModernInstrumentation(log, instrumentationConfig, callback);
                    }else{
                        log.info("Using obsolete instrumentation.")
                        slNodeCover.utils.runInstrumentation(instrumentationConfig, callback);
                    }


                });
                return promise;
            });
        }

        promiseChain = promiseChain.then(function () {
            return Promise.resolve();
        }).catch(function (err) {
            return Promise.reject(err);
        });

        return promiseChain;
    }

    this.run = function run(buildArguments) {
        return createBuildDiffPromiseChain(buildArguments);
    }

    function doModernInstrumentation(log, instrumentationConfig, callback){
        log.info("Using modern instrumentation (using InstanbulJS instead sl-node-cover).");
        var BrowserInstrumenter = require('./instrumentation/browser-instrumenter');
        var dvr =  new DefaultValuesResolver();
        instrumentationConfig.extensionsToInstrument = dvr.getFileExtensions();
        var bi = new BrowserInstrumenter(instrumentationConfig, log);
        bi.instrument(callback);
    }

    function loadBranchHistory(handlerParams, scmInstance) {
        return scmInstance.getBranchHistory(handlerParams, log).then(function (gitLog) {
            if (gitLog && gitLog.commitHistory && gitLog.commitHistory.length && gitLog.commitLog && gitLog.commitLog.length) {
                log.info('Branch history loaded (' + gitLog.commitHistory.length + ' commits)');
                handlerParams.finalOutput.meta.history = gitLog.commitHistory;
                handlerParams.finalOutput.meta.commitLog = gitLog.commitLog;
            } else {
                logBranchHistoryNotFound(gitLog, log);
            }
            return loadCommitIndexMap(gitLog, handlerParams, log);
        });
    }

    function getRepositoryUrl(handlerParams, scmInstance) {
        return scmInstance.getRepositoryUrl(handlerParams, log).then(function (repositoryUrl) {
            if (repositoryUrl) {
                log.info('Repository Url loaded (' + repositoryUrl + ')');
                handlerParams.finalOutput.meta.repositoryUrl = repositoryUrl;
            }
            return repositoryUrl;
        });
    }

    function getContributors(handlerParams, scmInstance) {
        return scmInstance.getContributors(handlerParams, log).then(function (contributorsLog) {
            if (contributorsLog && contributorsLog.contributors && contributorsLog.emailToIndexMap) {
                log.info('Contributors list loaded (' + contributorsLog.contributors.length + ' contributors)');
                handlerParams.finalOutput.meta.contributors = contributorsLog.contributors;
                handlerParams.emailToIndexMap = contributorsLog.emailToIndexMap;
            }else{
                logContributorsNotFound(contributorsLog, log);
            }
            return contributorsLog;
        })
    }

    function getCommitsPerFile(handlerParams, scmInstance){
        return scmInstance.getCommitsPerFile(handlerParams, log).then(function(fileToCommitsMap){
            if(fileToCommitsMap != null){
                log.info('File commit map loaded (' + Object.keys(fileToCommitsMap).length + ' records)');
                handlerParams.fileToCommitsMap = fileToCommitsMap;
            }else{
                log.error("Commits map is null or undefined.")
            }
            return fileToCommitsMap;
        })
    }
}


function logContributorsNotFound(contributorsLog, log) {
    var fieldsMap = {
        contributorsLog: !contributorsLog,
        contributors: contributorsLog && !contributorsLog.contributors,
        emailToIndexMap: contributorsLog && !contributorsLog.emailToIndexMap
    };
    var missing = utils.getMissingProperties(fieldsMap)
    log.error("Contributors list not loaded, missing fields: " + missing);
}

function logBranchHistoryNotFound(gitLog, log) {
    var fieldsMap = {
        gitLog: !gitLog,
        commitHistory: gitLog && !gitLog.commitHistory,
        commitHistoryLength: gitLog && gitLog.commitHistory && !gitLog.commitHistory.length,
        commitLog: gitLog && !gitLog.commitLog,
        commitLogLength: gitLog && gitLog.commitLog && !gitLog.commitLog.length
    };
    var missing = utils.getMissingProperties(fieldsMap)
    log.error("Branch history not loaded, missing fields: " + missing);
}

function clone(o) {
    //if (!o) return o; //Unreachable
    return JSON.parse(JSON.stringify(o));
}

var getProcessedFiles =function (workspacepath, processedFiles, logger) {
   if(CiaEnvVars.isEnableModernInstrumentation()){
       return fileAndFolderUtils.scanDirRecursively(workspacepath, logger)
   }
   return processedFiles;
}
var loadCommitIndexMap = function (gitLog, handlerParams, log) {
    if(!handlerParams.buildArguments.sendContributors){
        return gitLog;
    }
    if (gitLog && gitLog.commitToIndexMap) {
        handlerParams.commitToIndexMap = gitLog.commitToIndexMap;
    } else {
        log.error("Commit log index map not loaded")
    }
    return gitLog;
};

module.exports.BuildDiffProcess = BuildDiffProcess;
module.exports.sourceControlProviders = require("./source-control-providers.js");
