﻿var traverse = require('ast-traverse'),
    astUtils = require('sl-esprima-ast-utils'),
    md5 = require('md5'),
    utils = require('../utils'),
    guessMethodNameEspirma = require('./guess-method-name-espirma').guessMethodName,
    guessMethodNameBabylon = require('./guess-method-name-babylon').guessMethodName,
    consoleFileLogger = require("../sl-logger.js").consoleFileLogger,
    babelGenerator = require("babel-generator"),
    globalErrorHandler = require('../global-error-handler'),
    babylon = require("babylon"),
    SourceMapHandler = require("./source-map-handler").SourceMapHandler,
    createNodeKey = require('./create-node-key').createNodeKey,
    guessMethodLocation= require('./guess-method-location').guessMethodLocation;

var branchNodeTypes = {
    "IfStatement": true,
    "SwitchStatement": true,
    "ConditionalExpression": true,
    "LogicalExpression": true
};

function formatLocToArray(loc) {
    return [loc.line, loc.column];
}

// Brake the expression to pieces (every "or" will be different branch).
function findLogicalExpressionLeaves(node, leaves) {
    if (node.type === "LogicalExpression") {
        findLogicalExpressionLeaves(node.left, leaves);
        findLogicalExpressionLeaves(node.right, leaves);
    } else {
        leaves.push(node);
    }
}

// Calculate hash for function signature.
function calculateFunctionNodeSigHash(node, opts) {
    if (!node.params || node.params.length == 0) return "";
    var paramsStr = node.params.map(function (param) {
        return generateJsCodeFromAst(param, opts);
    }).join(',')
    var hash = md5(paramsStr);
    return hash;
}

function calculateFunctionNodeHash(node, isBranchCoverage, opts) {
    var clonedNode = astUtils.clone(node);
    var branchesData = [];
    
    // Visit node in pre order and calculate body hash.
    traverse(clonedNode, {
        pre: function (node, parent, prop, idx) {
            // If current node is the root node given return (we want to calculate the body hash only).
            if (node == clonedNode) return;

            // When we have inner function we set it to empty block statement, because we want to calculate the hash and diff for outer function only.
            // ObjectMethod is property function like "get", "set", "init". that uses in babylon. (in espirma its called Property). 
            if (node.type == "FunctionDeclaration" || node.type == "FunctionExpression" || node.type == "ArrowFunctionExpression" || node.type == "ObjectMethod" ){
                node.body = {
                    type: "BlockStatement",
                    body: []
                };
                node.params = [];
                node.id = {
                    name: '_',
                    type : "Identifier"
                };
            }
            if (isBranchCoverage) {
                if (branchNodeTypes[node.type] && node.type !== "LogicalExpression") {
                    addBranchesData(node, branchesData, opts);
                }

                // If this node its not logical exp inside logical exp.
                if (node.type == "LogicalExpression" && (parent == undefined || parent.type !== "LogicalExpression")) {
                    addBranchesData(node, branchesData, opts);
                }
            }
        }
    });

    var generated = generateJsCodeFromAst(clonedNode, opts);
    var hash = md5(generated);
    if (isBranchCoverage) {
        return {
            hash: hash,
            branchesData: branchesData
        }
    }
    return {
        hash: hash
    }
}

/**
 * Taken from eslint - https://github.com/eslint/eslint/blob/183def6115cad6f17c82ef1c1a245eb22d0bee83/lib/eslint.js#L800
 */
function trimShebang(text) {
    return text.replace(/^#!([^\r\n]+)/, function (match, captured) {
        return "//" + captured;
    });
}

function addBranchesData(node, branchesDataArray, opts) {
    switch (node.type) {

        // Those types contains some kind of "if else" structure.
        case "IfStatement":
        case "ConditionalExpression":
            var branchIndex = 0;

            // consequent mean when condition equals true.
            if (node.consequent !== null) {
                branchesDataArray.push(createBranchData(node, node.consequent, node.type, branchIndex++, opts));
            }

            // altrnate equivalent to "else".
            if (node.alternate == null) {
                // Istanbul counts empty alternates and takes the consequent's location
                var clonedNode = astUtils.clone(node);
                clonedNode.alternate = {
                    type: "BlockStatement",
                    body: []
                };

                // Set alternate location to consequent location (same as istanbul).
                clonedNode.alternate.loc = node.consequent !== null ? node.consequent.loc : node.loc;
                branchesDataArray.push(createBranchData(clonedNode, clonedNode.alternate, node.type, branchIndex++, opts));
            } else {
                branchesDataArray.push(createBranchData(node, node.alternate, node.type, branchIndex++, opts));
            }
            break;
        case "SwitchStatement":
            var branchIndex = 0;

            // Create branch from each case statement.
            node.cases.forEach(function (caseNode) {
                if (caseNode !== null) {
                    branchesDataArray.push(createBranchData(node, caseNode, node.type, branchIndex++, opts));
                }
            });
            break;
        case "LogicalExpression":
            var branchIndex = 0;
            var leaves = [];
            findLogicalExpressionLeaves(node, leaves);
            leaves.forEach(function (leafNode) {
                if (leafNode !== null) {
                    branchesDataArray.push(createBranchData(node, leafNode, node.type, branchIndex++, opts));
                }
            });
            break;
    }
}

function createBranchData(parentNode, node, nodeType, branchIndex, opts) {
    var clonedNode = astUtils.clone(node);
    var position = formatLocToArray(node.loc.start);
    var endPosition = formatLocToArray(node.loc.end);
    //In all cases except "ConditionalExpression" Istanbul reports in runtime on the location
    //of the parent node.
    var parentNodeStart = (nodeType === "ConditionalExpression" || nodeType === "LogicalExpression" || nodeType === "SwitchStatement") ? node.loc.start : parentNode.loc.start;
    var parentPosition = formatLocToArray(parentNodeStart)
    traverse(clonedNode, {
        pre: function (node, parent, prop, idx) {
            if (node == clonedNode) return;

            // If node cotains inner function set it to empty block statement, because we want to calculate hash (for diff) only for outer function.
            // ObjectMethod is property function like "get", "set", "init". that uses in babylon. (in espirma its called Property).
            if (node.type == "FunctionDeclaration" || node.type == "FunctionExpression" || node.type == "ArrowFunctionExpression" || node.type == "ObjectMethod") {
                node.body = {
                    type: "BlockStatement",
                    body: []
                };
                node.params = [];
                node.id = {
                    name: '_',
                    type : "Identifier"
                };
            }

            // Also empty inner branches.
            switch (node.type) {
                case "IfStatement":
                case "ConditionalExpression":
                    node.consequent = {
                        type: "BlockStatement",
                        body: []
                    };
                    node.alternate = {
                        type: "BlockStatement",
                        body: []
                    };
                    break;
                case "SwitchStatement":
                    node.cases = [];
                    break;
                case "LogicalExpression":
                    node.operator = "";
                    node.left = {
                        type: opts && opts.useBabylon ? "StringLiteral": "Literal",
                        value: null
                    };
                    node.right = {
                        type: opts && opts.useBabylon ? "StringLiteral": "Literal",
                        value: null
                    };
                    break;
            }
        }
    });
    var generated =  generateJsCodeFromAst(clonedNode, opts);
    var hash = md5(generated);
    return {
        type: nodeType,
        position: position,
        endPosition: endPosition,
        parentPosition: parentPosition,
        hash: hash,
        index:branchIndex
    };
}







function replaceWithOriginalPosition(position, sourceMaps) {
    var originalPosition = sourceMaps.originalPositionFor({
        line: position[0],
        column: position[1]
    });
    if (originalPosition && originalPosition.line !== null && originalPosition.column !== null) {
        return formatLocToArray(originalPosition);
    }
    return position;
}

function updateToSourceData(handler, method, node) {
    var sourceName = handler.getOriginalName();
    if(sourceName || method.name){
        updateMethodName(sourceName, method, node)
    }
    fillOriginalData(method);
    method.uniqueName = handler.getUniqueName();
    method.filename = handler.fixedOriginalFilename;
    method.start = handler.start.asFormatArray();
    method.end = handler.end.asFormatArray();
    method.name = handler.getOriginalName();
}

function fillOriginalData(method) {
    method.originalFilename = method.filename;
    method.originalStart = method.start;
    method.originalEnd = method.end;
    method.originalName = method.name;
}

function updateMethodName(sourceName, method, node) {
        method.originalName = method.name;
        method.name = sourceName;
        if (!node.id)
            node.id = {
                name: sourceName
            };
        else if (!node.id.name)
            node.id.name = sourceName;
}




function generateFileSignature(filename, relativeFileName, source, sourceMaps, isBranchCoverage, opts, logger) {
    var result = {
        module: {
            hash: ""
        },
        methods: [],
        branches: []
    };
    var log = logger ? logger.child({
        methodName: "generateFileSignature"
    }) : consoleFileLogger("generateFileSignature");

    filename = utils.adjustPathSlashes(filename); //we want all transients to use unix / instead of windows \

    var ast;

    try {
        source = trimShebang(source);

        //1. Parse the file
        ast = createAstFromFile(source, opts);

    } catch (parseException) {
        //TODO: Test
        globalErrorHandler.setLastError(parseException, "Failed to parse AST of " + filename + ". ");
        return {
            error: "Parse error (1)"
        };
    }

    try {
        var nodeToParent = {};
        // print AST node types, pre-order (node first, then its children) 
        traverse(ast, {
            pre: function (node, parent, prop, idx) {
                var key = createNodeKey(node);
                nodeToParent[key] = parent;
                // ObjectMethod is property function like "get", "set", "init". that uses in babylon. (in espirma its called Property).
                if (node.type == "FunctionDeclaration" || node.type == "FunctionExpression" || node.type == "ArrowFunctionExpression" || node.type == "ObjectMethod" || node.type == "ClassMethod") {
                    node.isFunctionNode = true;
                    //Global functions and iterators have different location and we need to get the location from the "id" in order to match the footprints.
                    var location = guessMethodLocation(node, nodeToParent);

                    var methodHashObject = calculateFunctionNodeHash(node, isBranchCoverage , opts),
                        hash = methodHashObject.hash, 
                        sigHash = calculateFunctionNodeSigHash(node, opts),
                        start = formatLocToArray(location.start),
                        end = formatLocToArray(node.loc.end),
                        nameAndType = getNameAndType(node, parent, opts),
                        name = nameAndType.name, //the actual name, the guessed name, or "(Anonymous)"
                        type = nameAndType.type,
                        isAnonymous = nameAndType.isAnonymous;

                    var uniqueName = [name, filename, start.join(',')].join('@');
                 
                    var method = {
                        name: name, //Method that can be presented to the user
                        isAnonymous: isAnonymous,
                        start: start,
                        end: end,
                        hash: hash,
                        sigHash: sigHash,
                        uniqueName: uniqueName,
                        filename: relativeFileName,
                        originalFilename: null,
                        originalStart: null,
                        originalEnd: null,
                        originalName: null,
                        type: type
                    };
                    
                    // If we have sourcemaps we need to get the location from original file.
                    if (sourceMaps) {
                        try {
                            var handler = new SourceMapHandler(sourceMaps, name, start, end, filename, log);
                            if(handler.shouldSkipped){return}
                            updateToSourceData(handler, method, node);
                        } catch (e) {
                            globalErrorHandler.setLastError(e, 'Error calculating source maps:');
                            log.error('Error calculating source maps:' + e);
                        }
                    }
                    if (isBranchCoverage) {
                        if (sourceMaps) {
                            methodHashObject.branchesData.forEach(function (branchData) {
                                branchData.position = replaceWithOriginalPosition(branchData.position, sourceMaps);
                                branchData.endPosition = replaceWithOriginalPosition(branchData.endPosition, sourceMaps);
                                branchData.parentPosition = replaceWithOriginalPosition(branchData.parentPosition, sourceMaps);
                                branchData.originalFilename = method.filename;
                            })
                        }

                        method.branches = methodHashObject.branchesData;      
                    }
                    result.methods.push(method);
                }
                if (node.hasOwnProperty("isFunctionNode") ||
                    (parent !== null && parent.hasOwnProperty("isFunctionNode"))) {
                    // We want to tag nodes that have a parent that is a method
                    node.isFunctionNode = true;
                }
                if (branchNodeTypes[node.type] &&
                    (!node.hasOwnProperty("isFunctionNode") || node.isFunctionNode !== true)) {
                    // Nodes that are of branch types and do not belong to a method parent are global branches
                    //TODO Global branches
                }
            }
        });
    } catch (traverseException1) {
        globalErrorHandler.setLastError(traverseException1, "Failed to traverse ast. Filename: " + filename + ". Error: ");
        return {
            error: 'Parse error(2)'
        }
    }

    try {
        var bodyHash = calculateFunctionNodeHash(ast, isBranchCoverage, opts).hash;
        result.module.hash = bodyHash;

    } catch (traverseException2) {
        globalErrorHandler.setLastError(traverseException2, "Failed to calculate module hash. Filename: " + filename );
        return {
            error: "Parse error (3)"
        };
    }

    return result;
}

var escodegen;
function generateJsCodeFromAst(ast, opts) {
    
    if (opts && opts.useBabylon) {
        return babelGenerator.default(ast, {
            comments: false
        }, '').code;
    }
    else{
        if (!escodegen){
            escodegen = require('escodegen')
        }
        return escodegen.generate(ast);
    }
}

var esprima;
function createAstFromFile(source, opts) {
    var sourceType = (opts && opts.es6Modules) ? 'module' : 'script';
    if (opts && opts.useBabylon) {   
        return babylon.parse(source, {
            // parse in strict mode and allow module declarations 
            sourceType: sourceType,
            plugins: [
                "*"
            ].concat(opts.babylonPlugins)
        })
    }
    else{
        if (!esprima){
            esprima = require('esprima');
        }
        return esprima.parse(source, {
            loc: true,
            range: true,
            comment: false, //No need for those,
            sourceType: sourceType
        });
    }

    

}

function getNameAndType(node, parent, opts){
    if(opts && opts.useBabylon)
        return guessMethodNameBabylon(node, parent);
    return guessMethodNameEspirma(node, parent);
}

module.exports.generateFileSignature = generateFileSignature;